// #include <zephyr/kernel.h>
// #include <zephyr/device.h>
// #include <zephyr/drivers/uart.h>
// #include <zephyr/logging/log.h>

// LOG_MODULE_REGISTER(NRF52_LOG,4);

// uint8_t tx_buffer[] = "Hi From nrf52840 SoC";
// uint8_t new_buffer[100];
// uint8_t buffer_rx[100];
// #define RING_BUFFER_SIZE 100

// uint8_t rx_buffer[RING_BUFFER_SIZE];
// static void uart_callback(const struct device *dev,void *user_data)
// {
//     // Gérer les événements UART ici
//     if(uart_irq_update(dev) && uart_irq_is_pending(dev))
//     {
//         if(uart_irq_rx_ready(dev))
//         {
//             uint8_t buffer[1];
//             uart_fifo_read(dev, buffer, 1);
//             printk("Caractère reçu: %c\n", buffer[0]);
//         }
//         if(uart_irq_tx_ready(dev))
//         {
//             // Gérer les événements de transmission ici
//             int ret = uart_fifo_fill(dev, "Hello World!", 12);
//             if(ret <= 0)
//             {
//                 printk("Erreur de transmission : err : %d \n",ret);
//             }else{
//                 printk("%d octets transmis \n",ret);
//             }

//             while (uart_irq_tx_complete(dev) == 0);
//             printk("Transmission terminée\n");
//             uart_irq_tx_disable(dev);
//         }
//     }
// }

// void uart_cb(const struct device *uart_dev,
// 							 struct uart_event *evt, void *user_data)
// {

// 	switch (evt->type) {
// 		case UART_TX_DONE:
//             LOG_DBG("Transmission terminée\n");
// 			break;
		
//         case UART_TX_ABORTED:
//             LOG_ERR("Transmission annulée\n");
// 		case UART_RX_RDY:
//             memset(buffer_rx,'\0',sizeof(buffer_rx));
//             for(int i = 0 ; i < evt->data.rx.len ; i++){
//                 buffer_rx[i] = evt->data.rx.buf[i + evt->data.rx.offset ];
//                 // LOG_DBG("buffer_rx[%d] : %c\n",i,buffer_rx[i]);
//             }
//             // memset(rx_buffer,'\0',sizeof(rx_buffer));
//             LOG_DBG("offset : %d\n",evt->data.rx.offset);
//             LOG_DBG("rx_buffer : %s , bytes : %d\n",evt->data.rx.buf,evt->data.rx.len);
//             LOG_DBG("rx_buffer : %s \n",rx_buffer);
//             LOG_DBG("buffer_rx : %s \n",buffer_rx);
// 			break;
		
// 		case UART_RX_BUF_REQUEST:
//             uart_rx_buf_rsp(uart_dev, new_buffer, sizeof(new_buffer));

//             //printk("new buffer is provided \n");
// 			break;

// 		case UART_RX_BUF_RELEASED:
// 			break;

// 		case UART_RX_DISABLED:
//             LOG_ERR("rx is disabled \n");
//             uart_rx_enable(uart_dev,rx_buffer,sizeof(rx_buffer),SYS_FOREVER_US);
// 			break;
		
// 		default:
// 			break;
// 	}
// }
// int main(void)
// {
//     const struct device *uart_dev = DEVICE_DT_GET(DT_NODELABEL(uart1));

//     if (!uart_dev) {
//         LOG_ERR("Impossible de trouver le périphérique UART\n");
//         return;
//     }
//     if(device_is_ready(uart_dev) == false)
//     {
//         LOG_ERR("Périphérique UART non prêt\n");
//         return -1;
//     }
//     // to use interrupt mode 
//     /*uart_irq_callback_set(uart_dev, &uart_callback);
//     uart_irq_rx_enable(uart_dev);

//     while (1) {
//         // Votre code principal ici
//         uart_irq_tx_enable(uart_dev);
//         k_sleep(K_MSEC(6000));
//     }*/
//     // Initialisez le buffer circulaire

//     int ret = uart_callback_set(uart_dev,&uart_cb,NULL);
//     if(ret < 0)
//     {
//         LOG_ERR("Erreur de configuration de la callback , ret : %d\n",ret);
//         return -1;
//     }
//     ret = uart_rx_enable(uart_dev,rx_buffer,sizeof(rx_buffer),1000);
//     if(ret < 0)
//     {
//         LOG_ERR("Erreur de configuration de la réception , ret : %d\n",ret);
//         return -1;
//     }
//     while (true)
//     {
//         ret = uart_tx(uart_dev,tx_buffer,sizeof(tx_buffer),0);
//         if(ret < 0)
//         {
//             LOG_ERR("Erreur de transmission , ret : %d\n",ret);
//         }
//         k_sleep(K_MSEC(5000));
//         uart_tx(uart_dev,"did you received the packet",sizeof("did you received the packet"),0);
//         k_sleep(K_MSEC(2000));
//     }
    
//     return 0;
// }

/*
 * Copyright (c) 2016 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/sys/ring_buffer.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/uart.h>
#include <string.h>
#include <stdio.h>
#include <zephyr/logging/log.h>

LOG_MODULE_REGISTER(NRF52_LOG,4);
#define UART_BUF_SIZE		23
#define UART_TX_TIMEOUT_MS	100
#define UART_RX_TIMEOUT_MS	100

K_SEM_DEFINE(tx_done, 1, 1);
K_SEM_DEFINE(rx_disabled, 0, 1);

#define UART_TX_BUF_SIZE  		256
#define UART_RX_MSG_QUEUE_SIZE	8
struct uart_msg_queue_item {
	uint8_t bytes[UART_BUF_SIZE];
	uint32_t length;
};

// UART TX fifo
RING_BUF_DECLARE(app_tx_fifo, UART_TX_BUF_SIZE);
volatile int bytes_claimed;

// UART RX primary buffers
uint8_t uart_double_buffer[2][UART_BUF_SIZE];
uint8_t *uart_buf_next = uart_double_buffer[1];

// UART RX message queue
K_MSGQ_DEFINE(uart_rx_msgq, sizeof(struct uart_msg_queue_item), UART_RX_MSG_QUEUE_SIZE, 4);

static const struct device *dev_uart;

static int uart_tx_get_from_queue(void)
{
	uint8_t *data_ptr;
	// Try to claim any available bytes in the FIFO
	bytes_claimed = ring_buf_get_claim(&app_tx_fifo, &data_ptr, UART_TX_BUF_SIZE);

	if(bytes_claimed > 0) {
		// Start a UART transmission based on the number of available bytes
		uart_tx(dev_uart, data_ptr, bytes_claimed, SYS_FOREVER_MS);
	}
	return bytes_claimed;
}

void app_uart_async_callback(const struct device *uart_dev,
							 struct uart_event *evt, void *user_data)
{
	static struct uart_msg_queue_item new_message;

	switch (evt->type) {
		case UART_TX_DONE:
			// Free up the written bytes in the TX FIFO
			ring_buf_get_finish(&app_tx_fifo, bytes_claimed);

			// If there is more data in the TX fifo, start the transmission
			if(uart_tx_get_from_queue() == 0) {
				// Or release the semaphore if the TX fifo is empty
				k_sem_give(&tx_done);
			}
			break;
		
		case UART_RX_RDY:
			memcpy(new_message.bytes, evt->data.rx.buf + evt->data.rx.offset, evt->data.rx.len);
			new_message.length = evt->data.rx.len;
            LOG_DBG("new_message.length : %d\n",new_message.length);
			if(k_msgq_put(&uart_rx_msgq, &new_message, K_NO_WAIT) != 0){
				printk("Error: Uart RX message queue full!\n");
			}
			break;
		
		case UART_RX_BUF_REQUEST:
			uart_rx_buf_rsp(dev_uart, uart_buf_next, UART_BUF_SIZE);
			break;

		case UART_RX_BUF_RELEASED:
			uart_buf_next = evt->data.rx_buf.buf;
			break;

		case UART_RX_DISABLED:
			k_sem_give(&rx_disabled);
			break;
		
		default:
			break;
	}
}

static void app_uart_init(void)
{
	dev_uart = DEVICE_DT_GET(DT_NODELABEL(uart1));

    if (!dev_uart) {
        LOG_ERR("Impossible de trouver le périphérique UART\n");
        return;
    }
    if(device_is_ready(dev_uart) == false)
    {
        LOG_ERR("Périphérique UART non prêt\n");
        return -1;
    }

	uart_callback_set(dev_uart, app_uart_async_callback, NULL);
	uart_rx_enable(dev_uart, uart_double_buffer[0], UART_BUF_SIZE, UART_RX_TIMEOUT_MS);
}

// Function to send UART data, by writing it to a ring buffer (FIFO) in the application
// WARNING: This function is not thread safe! If you want to call this function from multiple threads a semaphore should be used
static int app_uart_send(const uint8_t * data_ptr, uint32_t data_len)
{
	while(1) {
		// Try to move the data into the TX ring buffer
		uint32_t written_to_buf = ring_buf_put(&app_tx_fifo, data_ptr, data_len);
		data_len -= written_to_buf;
		
		// In case the UART TX is idle, start transmission
		if(k_sem_take(&tx_done, K_NO_WAIT) == 0) {
			uart_tx_get_from_queue();
		}	
		
		// In case all the data was written, exit the loop
		if(data_len == 0) break;

		// In case some data is still to be written, sleep for some time and run the loop one more time
		k_msleep(10);
		data_ptr += written_to_buf;
	}

	return 0;
}

void main(void)
{
	printk("UART Async example started\n");
	
	app_uart_init();

	const uint8_t test_string[] = "Hello world through the UART async driver\r\n";
	// app_uart_send(test_string, strlen(test_string));

	struct uart_msg_queue_item incoming_message;
    static uint8_t string_buffer[UART_BUF_SIZE + 1] = {0};
    int buffer_length = 0;

    while (1) {
        // This function will not return until a new message is ready
        k_msgq_get(&uart_rx_msgq, &incoming_message, K_FOREVER);

        // Copy incoming_message.bytes to string_buffer
        for (int i = 0; i < incoming_message.length; i++) {
            string_buffer[buffer_length] = incoming_message.bytes[i];
            buffer_length++;
            if (incoming_message.bytes[i] == '\x1A') {
                // End of message detected
                LOG_DBG("End of message detected , string_buffer[%d] : %c \n", buffer_length - 1, string_buffer[buffer_length - 1]);
                buffer_length--; // Decrease buffer length to remove the last character
                string_buffer[buffer_length] = '\0'; // Null-terminate the string
                printk("RX : %s\n", string_buffer);
                memset(string_buffer, 0, sizeof(string_buffer)); // Clear the buffer
                buffer_length = 0; // Reset buffer length
                break;
            }
        }
    }
}